from typing import Any, Union

from apispec import BasePlugin, APISpec
from apispec.exceptions import DuplicateComponentNameError
from pydantic import BaseModel


class PydanticPlugin(BasePlugin):
    """
    An APISpec plugin that will register Pydantic models with APISpec.

    Uses the built-in schema export provided by Pydantic with minor modifications to provide
    the right structure for OpenAPI.

    Example:
        from pydantic import BaseModel

        class MyModel(BaseModel):
            hello: str
            world: str

        spec = APISPec(plugins=[PydanticPlugin()])
        spec.components.schema("MyModel", model=MyModel)
    """

    def schema_helper(
        self, name: str, definition: dict, **kwargs: Any
    ) -> Union[dict, None]:
        model: Union[BaseModel, None] = kwargs.pop("model", None)
        if model:
            schema = model.schema(ref_template="#/components/schemas/{model}")

            # If the spec has passed, we probably have nested models to contend with.
            spec: Union[APISpec, None] = kwargs.pop("spec", None)
            if spec and "definitions" in schema:
                for k, v in schema["definitions"].items():
                    try:
                        spec.components.schema(k, v)
                    except DuplicateComponentNameError:
                        pass

            if "definitions" in schema:
                del schema["definitions"]

            return schema

        return None



from typing import Type, Optional, Union, List

from apispec import APISpec
from pydantic import BaseModel

DEFAULT_DESCRIPTION = "Success"
DEFAULT_CODE = 200


class Response:
    def __init__(self, model: type, code: int = 200, description: str = "Success"):
        self.model = model
        self.code = code
        self.description = description


class Operation:
    def __init__(
        self,
        request: Optional[Type[BaseModel]] = None,
        response: Optional[Union[Response, Type[BaseModel]]] = None,
        responses: Optional[List[Response]] = None,
    ):
        self.request = request

        if response and responses:
            raise TypeError("You must only pass one of response or responses")

        if response:
            self._populate_response(response)
        else:
            self._populate_responses(responses)

    def _populate_response(self, response: Union[Response, type]):
        if isinstance(response, Response):
            # If this is a Response object, we can track it as-is.
            self.responses = {response.code: response}
        else:
            # If not, we will use sensible defaults
            self.responses = {DEFAULT_CODE: Response(model=response)}

    def _populate_responses(self, responses: List[Response]):
        self.responses = {}
        for response in responses:
            if response.code in self.responses:
                raise TypeError(
                    "You  must only specify one response per HTTP status code"
                )
            self.responses[response.code] = response


Method = Union[Type[BaseModel], Operation]


class Docs:
    methods = ["get", "post", "put", "patch", "delete", "head", "options"]

    def __init__(
        self,
        summary: str = None,
        get: Method = None,
        post: Method = None,
        put: Method = None,
        patch: Method = None,
        delete: Method = None,
        head: Method = None,
        options: Method = None,
        request: Type[BaseModel] = None,
        response: Union[Response, Type[BaseModel]] = None,
        responses: List[Response] = None,
    ):
        self.summary = summary

        self.request = request
        self.response = response
        self.responses = responses

        self.get = get
        self.post = post
        self.put = put
        self.patch = patch
        self.delete = delete
        self.head = head
        self.options = options

        # check to make sure we haven't violated the contract
        if self.request or self.response or self.responses:
            for method in self.methods:
                if getattr(self, method):
                    raise TypeError(
                        "You must choose either a short-hand or long-hand Docs, not both."
                    )

    @classmethod
    def _build_operation_from_operation(cls, method: Operation, spec: APISpec):
        operation = {}

        if method.request:
            if method.request.__name__ not in spec.components.schemas:
                spec.components.schema(
                    method.request.__name__, model=method.request, spec=spec
                )
            operation["requestBody"] = {
                "content": {
                    "application/json": {
                        "schema": method.request.__name__,
                    }
                }
            }

        if method.responses:
            responses = {}

            for code, response in method.responses.items():
                if response.model.__name__ not in spec.components.schemas:
                    spec.components.schema(
                        response.model.__name__,
                        model=response.model,
                        spec=spec,
                    )
                responses[code] = {
                    "description": response.description,
                    "content": {
                        "application/json": {
                            "schema": response.model.__name__,
                        }
                    },
                }

            operation["responses"] = responses

        return operation

    @classmethod
    def _build_operation_from_model(cls, model: Type[BaseModel], spec: APISpec):
        return cls._build_operation_from_operation(Operation(response=model), spec)

    @classmethod
    def _build_operation(cls, method: Method, spec: APISpec):
        if isinstance(method, Operation):
            return cls._build_operation_from_operation(method, spec)
        else:
            return cls._build_operation_from_model(method, spec)

    def _build_simple_operation(self, spec: APISpec):
        return self._build_operation_from_operation(
            Operation(
                request=self.request, response=self.response, responses=self.responses
            ),
            spec,
        )

    def build_operations(self, spec: APISpec, methods: List[str]):
        operations = {}

        if self.request or self.responses or self.response:
            if len(methods) != 1:
                raise TypeError(
                    "You can only use Docs short-hand for single-method API routes."
                )

            operations[methods[0].lower()] = self._build_simple_operation(spec)

        else:
            for method in self.methods:
                if getattr(self, method):
                    operations[method] = self._build_operation(
                        getattr(self, method), spec
                    )

        return operations


Resp = Response
Op = Operation



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chalice_spec']

package_data = \
{'': ['*']}

install_requires = \
['apispec>=6.0.2,<7.0.0']

setup_kwargs = {
    'name': 'chalice-spec',
    'version': '0.3.0',
    'description': 'Chalice x APISpec x Pydantic plug-ins',
    'long_description': '# chalice-spec\n\n[![Python package](https://github.com/TestBoxLab/chalice-spec/actions/workflows/test.yml/badge.svg)](https://github.com/TestBoxLab/chalice-spec/actions/workflows/test.yml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**Chalice × APISpec × Pydantic plug-ins**\n\nCombines the power of Chalice, APISpec, and Pydantic to make AWS Chalice apps easily documented\n\n## Installation\n\nFirst, add chalice-spec:\n\n```shell\npoetry add chalice_spec\n```\n\nWe consider Chalice, APISpec, and Pydantic "peer dependencies." We only include them as dev\ndependencies in our codebase, and you may need to install them in yours if you haven\'t\nalready.\n\n```shell\npoetry add chalice apispec pydantic\n```\n\n## Setup\n\nFirst, instantiate your APISpec object with *both* the Pydantic and Chalice plug-ins, assuming\nyou need the functionality of each. While the Pydantic plugin can be used alone, you currently must use\nthe Pydantic plugin with the Chalice plugin.\n\n```python\napp = Chalice(app_name="hello_world")\nspec = APISpec(chalice_app=app,\n               ...,\n               plugins=[PydanticPlugin(), ChalicePlugin()])\n```\n\nIf you use:\n```python\nChalicePlugin(generate_default_docs=True)\n```\nthe plugin will generate empty docs (with empty request and response schemas) for every endpoint that you\'ve defined in your app. This can be useful as a starting point / overview while developing.\n\n## Usage\n\nTo document your API, use your existing Pydantic models and add kwargs to Chalice decorators.\n\n**Before:**\n```python\n@app.route(\'/\', methods=["POST"])\ndef example():\n    body = MySchema.parse_obj(app.current_request.json_body)\n```\n\n**After:**\n```python\n@app.route(\'/\', methods=["POST"], docs=Docs(\n    post=Operation(request=MySchema)\n))\ndef example():\n    body = MySchema.parse_obj(app.current_request.json_body)\n```\n\nIf you have multiple methods supported, you may have something like:\n\n```python\n@app.route(\'/\', methods=["POST", "PUT"],\n           docs=Docs(\n               post=Operation(request=MyCreateSchema, response=MyReadSchema),\n               put=Operation(request=MyEditSchema, response=MyReadSchema)\n           )\ndef example():\n    # code goes here\n    pass\n```\n\n### API\n\n- [ ] TODO: this section coming soon!\n',
    'author': 'Jake Wood',
    'author_email': 'jake@testbox.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/TestBoxLab/chalice-spec',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from apispec import BasePlugin, APISpec
from pydantic import BaseModel

from chalice_spec.docs import Docs, Operation


class ChalicePlugin(BasePlugin):
    """
    An APISpec plugin which will monkeypatch Chalice in order to allow for very
    convenient API documentation. It is designed to work with in conjunction with
    the PydanticPlugin.

    For example...

    @app.route('/hello',
               post_body_model=APydanticModel,
               post_response_body=APydanticModel)
    def hello():
        return {"world": "The quick brown fox jumps over the lazy dog."}
    """

    def __init__(self, generate_default_docs: bool = False):
        super(ChalicePlugin, self).__init__()
        self._generate_default_docs = generate_default_docs

    def init_spec(self, spec: APISpec) -> None:
        """
        When we initialize the spec, we should also monkeypatch the Chalice app
        we are working with.

        :param spec: APISpec object to work with
        :return: None
        """
        chalice_app = spec.options.pop("chalice_app")

        original_route = chalice_app.route

        def route(path: str, **kwargs):
            """
            Register a new route on a Chalice app. Monekypatched to support APISpec instructions.
            :param path: The path to use.
            :param methods: the allowable methods and their documentation
            :param kwargs: Additional Chalice kwargs and APIspec definitions.
            """
            docs: Docs = kwargs.pop("docs", None)
            methods = [method.lower() for method in kwargs.get("methods", ["get"])]

            if docs is None and self._generate_default_docs:
                docs = Docs(
                    **{
                        method: Operation(
                            response=BaseModel,
                            request=(
                                None
                                if method in ["get", "delete", "head", "options"]
                                else BaseModel
                            ),
                        )
                        for method in methods
                    }
                )

            if docs:
                operations = docs.build_operations(spec, methods)
                spec.path(path, operations=operations, summary=docs.summary)

            return original_route(path, **kwargs)

        chalice_app.route = route
